#!/usr/bin/env python


class IgnoreThisItem(Exception):
    pass


class InvalidCallbackResult(Exception):
    pass


class InvalidFuncType(Exception):
    pass


class InvalidRequestMethod(Exception):
    pass


class NotImplementedParseError(Exception):
    pass


class NothingMatchedError(Exception):
    pass


class SpiderHookError(Exception):
    pass
